package com.biganto.visual.roompark.data.memcache

import androidx.lifecycle.Lifecycle
import androidx.lifecycle.LifecycleObserver
import androidx.lifecycle.OnLifecycleEvent
import androidx.lifecycle.ProcessLifecycleOwner
import com.jakewharton.rxrelay2.Relay
import com.jakewharton.rxrelay2.ReplayRelay
import io.reactivex.Observable
import io.reactivex.disposables.CompositeDisposable
import timber.log.Timber
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.TimeUnit

/**
 * Created by Vladislav Bogdashkin on 19.10.2018.
 */


/**
 * @param[maxSize] max entries count before automatic save && flush
 * @param[flushInterval] time without any actions with cache in SECONDS before automatic save && flush
 * */
abstract class EntityCache<K,V>

constructor(
    private val maxSize: Int = 300,
    private val flushInterval: Long = TimeUnit.MINUTES.toSeconds(1)

) : Cache<K,V> {

    private val disposable = CompositeDisposable()
    protected abstract fun saveDelegate  (values:List<V>)
    protected abstract fun readDelegate  (key:K) : V?
    protected abstract fun deleteDelegate  (value:V)

    protected val notifier: Relay<V> = ReplayRelay.create<V>().toSerialized()

    init {
        disposable.addAll(
            Observable.interval(flushInterval, flushInterval, TimeUnit.SECONDS)
                .doOnTerminate { clear() }
                .doOnDispose { clear() }
                .subscribe { clear() }
        )
    }

    override fun contains(key: K): Boolean = keyMap.containsKey(key)

    protected val locker = Any()

    private var lastFlushTime = System.nanoTime()

    private val keyMap = ConcurrentHashMap<K, V>()

    override val size: Int
        get() = keyMap.size

    override val toList
        get() = keyMap.toList()

    fun deleteEntity(key: K){
        synchronized(locker) {
            deleteItem(key)
        }
    }

    private fun deleteItem(key:K){
        synchronized(locker) {
            lastFlushTime = System.nanoTime()
            var toDelete = keyMap.remove(key)
            if (toDelete == null)
                toDelete = readDelegate(key)
            if (toDelete == null) return // -> нет в хранилище
            deleteDelegate(value = toDelete)
        }
    }

    fun deleteEntitys(keys: List<K>){
        synchronized(locker) {
            keys.forEach { deleteItem(it)}
        }
    }

    override fun set(key: K, value: V?) {
        synchronized(locker) {
            lastFlushTime = System.nanoTime()
            if (keyMap.size > maxSize)
                clear()
            value?.let {
                keyMap[key]=it
                notifier.accept(it)
            }
        }
    }

    override fun remove(key: K) = keyMap.remove(key)

    override fun get(key: K): V? {
        synchronized(locker) {
            lastFlushTime = System.nanoTime()
            return keyMap[key] ?: readDelegate(key)
        }
    }


    override fun saveAll(){
        synchronized(locker)
        {
            if (keyMap.size > 0) {
                Timber.d("Going to save items: ${keyMap.values.size}")
                saveDelegate(keyMap.values.toList())
            }
        }
    }

    override fun clear() {
        if (keyMap.size == 0) return
        synchronized(locker)
        {
            saveAll()
            keyMap.clear()
        }
    }


    override fun removeAll(): List<V>? {
        val values= keyMap.values.toList()
        keyMap.clear()
        return values
    }


    private fun recycle() {
        if (keyMap.size<=0) return
        val shouldRecycle = System.nanoTime() - lastFlushTime >= TimeUnit.MILLISECONDS.toNanos(flushInterval)
        if (!shouldRecycle) return
        keyMap.clear()
    }


}

interface Cache<K,V> {
    val size: Int

    val toList:List<Pair<K,V>>

    operator fun set(key: K, value: V?)

    fun contains(key: K): Boolean

    operator fun get(key: K): V?

    fun remove(key: K): V?

    fun removeAll(): List<V>?

    fun clear()
    fun saveAll()
}


/**
 * Suppress warning Leaking This as we sure to make singletone instance of object in single-thread
 * more info and
 * @see <a href="https://stackoverflow.com/questions/3921616/leaking-this-in-constructor-warning">discussion</a>
 */
@Suppress("LeakingThis")
abstract class LifeCycleCache<K,V>(size:Int, secondsToFlush:Long)
    : EntityCache<K, V>(maxSize=size,flushInterval = secondsToFlush)
    , LifecycleObserver
{
    init {
        ProcessLifecycleOwner.get().lifecycle
            .addObserver(this)
    }

    @OnLifecycleEvent(Lifecycle.Event.ON_PAUSE)
    fun onAppPaused() {
        saveAll()
    }
}
